@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2020 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -84,7 +84,7 @@ class CommandConfig(BaseConfig):
                            default=self.config)
         group.add_argument('--hooks', metavar='FILE', help='use FILE as hooks.py module',
                            default=self.hooks)
-        group.add_argument('--cache', metavar='FILE', help='use FILE as cache database',
+        group.add_argument('--cache', metavar='FILE', help='use FILE as cache database, alternatively can accept a redis URI',
                            default=self.cache)
 
         group = parser.add_argument_group('Authentication')
