@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2020 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -35,10 +35,12 @@ import os
 import re
 import subprocess
 import requests
+import textwrap
 import urlwatch
 from requests.packages.urllib3.exceptions import InsecureRequestWarning
 
 from .util import TrackSubClasses
+from .filters import FilterBase
 
 requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
 
@@ -85,12 +87,16 @@ class JobBase(object, metaclass=TrackSubClasses):
     def job_documentation(cls):
         result = []
         for sc in TrackSubClasses.sorted_by_kind(cls):
-            result.extend((
-                '  * %s - %s' % (sc.__kind__, sc.__doc__),
-                '    Required keys: %s' % (', '.join(sc.__required__),),
-                '    Optional keys: %s' % (', '.join(sc.__optional__),),
-                '',
-            ))
+            if sc.__doc__:
+                result.append('  * %s - %s' % (sc.__kind__, sc.__doc__))
+            else:
+                result.append('  * %s' % (sc.__kind__,))
+
+            for msg, value in (('    Required keys: ', sc.__required__), ('    Optional keys: ', sc.__optional__)):
+                if value:
+                    values = ('\n' + (len(msg) * ' ')).join(textwrap.wrap(', '.join(value), 79 - len(msg)))
+                    result.append('%s%s' % (msg, values))
+            result.append('')
         return '\n'.join(result)
 
     def get_location(self):
@@ -166,7 +172,7 @@ class JobBase(object, metaclass=TrackSubClasses):
 
 class Job(JobBase):
     __required__ = ()
-    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool')
+    __optional__ = ('name', 'filter', 'max_tries', 'diff_tool', 'compared_versions')
 
     # determine if hyperlink "a" tag is used in HtmlReporter
     LOCATION_IS_URL = False
@@ -193,6 +199,12 @@ class ShellJob(Job):
         if result != 0:
             raise ShellError(result)
 
+        # If the first filter is a bytes filter, return content in bytes instead of in unicode
+        # as that's what's required by the library used by that filter
+        if self.filter and (FilterBase.is_bytes_filter(self.filter)
+                            or FilterBase.is_bytes_filter(next(iter(self.filter[0])))):
+            return stdout_data
+
         return stdout_data.decode('utf-8')
 
 
@@ -203,7 +215,8 @@ class UrlJob(Job):
 
     __required__ = ('url',)
     __optional__ = ('cookies', 'data', 'method', 'ssl_no_verify', 'ignore_cached', 'http_proxy', 'https_proxy',
-                    'headers', 'ignore_connection_errors', 'ignore_http_error_codes', 'encoding', 'timeout')
+                    'headers', 'ignore_connection_errors', 'ignore_http_error_codes', 'encoding', 'timeout',
+                    'ignore_timeout_errors', 'ignore_too_many_redirects')
 
     LOCATION_IS_URL = True
     CHARSET_RE = re.compile('text/(html|plain); charset=([^;]*)')
@@ -278,6 +291,12 @@ class UrlJob(Job):
         # Save ETag from response into job_state, which will be saved in cache
         job_state.etag = response.headers.get('ETag')
 
+        # If the first filter is a bytes filter, return content in bytes instead of in unicode
+        # as that's what's required by the library used by that filter
+        if self.filter and (FilterBase.is_bytes_filter(self.filter)
+                            or FilterBase.is_bytes_filter(next(iter(self.filter[0])))):
+            return response.content
+
         # If we can't find the encoding in the headers, requests gets all
         # old-RFC-y and assumes ISO-8859-1 instead of UTF-8. Use the old
         # urlwatch behavior and try UTF-8 decoding first.
@@ -319,6 +338,10 @@ class UrlJob(Job):
     def ignore_error(self, exception):
         if isinstance(exception, requests.exceptions.ConnectionError) and self.ignore_connection_errors:
             return True
+        if isinstance(exception, requests.exceptions.Timeout) and self.ignore_timeout_errors:
+            return True
+        if isinstance(exception, requests.exceptions.TooManyRedirects) and self.ignore_too_many_redirects:
+            return True
         elif isinstance(exception, requests.exceptions.HTTPError):
             status_code = exception.response.status_code
             ignored_codes = []
