@@ -1,6 +1,6 @@
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2020 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -28,16 +28,13 @@
 
 
 import difflib
-import tempfile
-import subprocess
 import re
-import shlex
 import email.utils
 import itertools
 import logging
 import sys
 import time
-import cgi
+import html
 import functools
 
 import requests
@@ -57,8 +54,18 @@ try:
 except ImportError:
     Pushbullet = None
 
-logger = logging.getLogger(__name__)
+try:
+    import matrix_client.api
+except ImportError:
+    matrix_client = None
 
+try:
+    # markdown2 is an optional dependency which provides better formatting for Matrix.
+    from markdown2 import Markdown
+except ImportError:
+    Markdown = None
+
+logger = logging.getLogger(__name__)
 
 # Regular expressions that match the added/removed markers of GNU wdiff output
 WDIFF_ADDED_RE = r'[{][+].*?[+][}]'
@@ -107,28 +114,6 @@ class ReporterBase(object, metaclass=TrackSubClasses):
     def submit(self):
         raise NotImplementedError()
 
-    def unified_diff(self, job_state):
-        if job_state.job.diff_tool is not None:
-            with tempfile.NamedTemporaryFile() as old_file, tempfile.NamedTemporaryFile() as new_file:
-                old_file.write(job_state.old_data.encode('utf-8'))
-                old_file.flush()
-                new_file.write(job_state.new_data.encode('utf-8'))
-                new_file.flush()
-                cmdline = shlex.split(job_state.job.diff_tool) + [old_file.name, new_file.name]
-                proc = subprocess.Popen(cmdline, stdout=subprocess.PIPE)
-                stdout, _ = proc.communicate()
-                # Diff tools return 0 for "nothing changed" or 1 for "files differ", anything else is an error
-                if proc.returncode in (0, 1):
-                    return stdout.decode('utf-8')
-                else:
-                    raise subprocess.CalledProcessError(result, cmdline)
-
-        timestamp_old = email.utils.formatdate(job_state.timestamp, localtime=1)
-        timestamp_new = email.utils.formatdate(time.time(), localtime=1)
-        return ''.join(difflib.unified_diff([l + '\n' for l in job_state.old_data.splitlines()],
-                                            [l + '\n' for l in job_state.new_data.splitlines()],
-                                            '@', '@', timestamp_old, timestamp_new))
-
 
 class SafeHtml(object):
     def __init__(self, s):
@@ -138,8 +123,8 @@ class SafeHtml(object):
         return self.s
 
     def format(self, *args, **kwargs):
-        return str(self).format(*(cgi.escape(str(arg)) for arg in args),
-                                **{k: cgi.escape(str(v)) for k, v in kwargs.items()})
+        return str(self).format(*(html.escape(str(arg)) for arg in args),
+                                **{k: html.escape(str(v)) for k, v in kwargs.items()})
 
 
 class HtmlReporter(ReporterBase):
@@ -153,6 +138,7 @@ class HtmlReporter(ReporterBase):
         <html><head>
             <title>urlwatch</title>
             <meta http-equiv="content-type" content="text/html; charset=utf-8">
+            <meta name="viewport" content="width=device-width, initial-scale=1.0">
             <style type="text/css">
                 body { font-family: sans-serif; }
                 .diff_add { color: green; background-color: lightgreen; }
@@ -219,15 +205,16 @@ class HtmlReporter(ReporterBase):
             return SafeHtml('...')
 
         if difftype == 'table':
-            timestamp_old = email.utils.formatdate(job_state.timestamp, localtime=1)
-            timestamp_new = email.utils.formatdate(time.time(), localtime=1)
+            timestamp_old = email.utils.formatdate(job_state.timestamp, localtime=True)
+            timestamp_new = email.utils.formatdate(time.time(), localtime=True)
             html_diff = difflib.HtmlDiff()
-            return SafeHtml(html_diff.make_table(job_state.old_data.splitlines(1), job_state.new_data.splitlines(1),
+            return SafeHtml(html_diff.make_table(job_state.old_data.splitlines(keepends=True),
+                                                 job_state.new_data.splitlines(keepends=True),
                                                  timestamp_old, timestamp_new, True, 3))
         elif difftype == 'unified':
             return ''.join((
                 '<pre>',
-                '\n'.join(self._diff_to_html(self.unified_diff(job_state))),
+                '\n'.join(self._diff_to_html(job_state.get_diff())),
                 '</pre>',
             ))
         else:
@@ -246,7 +233,7 @@ class TextReporter(ReporterBase):
                 pretty_name = job_state.job.pretty_name()
                 location = job_state.job.get_location()
                 if pretty_name != location:
-                    location = '%s (%s)' % (pretty_name, location)
+                    location = '%s ( %s )' % (pretty_name, location)
                 yield ': '.join((job_state.verb.upper(), location))
             return
 
@@ -258,12 +245,12 @@ class TextReporter(ReporterBase):
             details.extend(details_part)
 
         if summary:
-            sep = line_length * '='
-            yield from itertools.chain(
+            sep = (line_length * '=') or None
+            yield from (part for part in itertools.chain(
                 (sep,),
                 ('%02d. %s' % (idx + 1, line) for idx, line in enumerate(summary)),
                 (sep, ''),
-            )
+            ) if part is not None)
 
         if show_details:
             yield from details
@@ -284,7 +271,7 @@ class TextReporter(ReporterBase):
         if job_state.old_data in (None, job_state.new_data):
             return None
 
-        return self.unified_diff(job_state)
+        return job_state.get_diff()
 
     def _format_output(self, job_state, line_length):
         summary_part = []
@@ -293,7 +280,7 @@ class TextReporter(ReporterBase):
         pretty_name = job_state.job.pretty_name()
         location = job_state.job.get_location()
         if pretty_name != location:
-            location = '%s (%s)' % (pretty_name, location)
+            location = '%s ( %s )' % (pretty_name, location)
 
         pretty_summary = ': '.join((job_state.verb.upper(), pretty_name))
         summary = ': '.join((job_state.verb.upper(), location))
@@ -301,11 +288,12 @@ class TextReporter(ReporterBase):
 
         summary_part.append(pretty_summary)
 
-        sep = line_length * '-'
+        sep = (line_length * '-') or None
         details_part.extend((sep, summary, sep))
         if content is not None:
             details_part.extend((content, sep))
-        details_part.extend(('', ''))
+        details_part.extend(('', '') if sep else ('',))
+        details_part = [part for part in details_part if part is not None]
 
         return summary_part, details_part
 
@@ -348,7 +336,7 @@ class StdoutReporter(TextReporter):
         cfg = self.report.config['report']['text']
         line_length = cfg['line_length']
 
-        separators = (line_length * '=', line_length * '-', '-- ')
+        separators = (line_length * '=', line_length * '-', '-- ') if line_length else ()
         body = '\n'.join(super().submit())
 
         for line in body.splitlines():
@@ -393,9 +381,14 @@ class EMailReporter(TextReporter):
             logger.debug('Not sending e-mail (no changes)')
             return
         if self.config['method'] == "smtp":
-            smtp_user = self.config['smtp'].get('user', self.config['from'])
+            smtp_user = self.config['smtp'].get('user', None) or self.config['from']
+            # Legacy support: The current smtp "auth" setting was previously called "keyring"
+            if 'keyring' in self.config['smtp']:
+                logger.info('The SMTP config key "keyring" is now called "auth".')
+            use_auth = self.config['smtp'].get('auth', self.config['smtp'].get('keyring', False))
             mailer = SMTPMailer(smtp_user, self.config['smtp']['host'], self.config['smtp']['port'],
-                                self.config['smtp']['starttls'], self.config['smtp']['keyring'])
+                                self.config['smtp']['starttls'], use_auth,
+                                self.config['smtp'].get('insecure_password'))
         elif self.config['method'] == "sendmail":
             mailer = SendmailMailer(self.config['sendmail']['path'])
         else:
@@ -432,7 +425,7 @@ class WebServiceReporter(TextReporter):
 
         try:
             service = self.web_service_get()
-        except Exception as e:
+        except Exception:
             logger.error('Failed to load or connect to %s - are the dependencies installed and configured?',
                          self.__kind__, exc_info=True)
             return
@@ -446,13 +439,25 @@ class PushoverReport(WebServiceReporter):
     __kind__ = 'pushover'
 
     def web_service_get(self):
+        if chump is None:
+            raise ImportError('Python module "chump" not installed')
+
         app = chump.Application(self.config['app'])
         return app.get_user(self.config['user'])
 
     def web_service_submit(self, service, title, body):
         sound = self.config['sound']
-        device = self.config['device']
-        msg = service.create_message(title=title, message=body, html=True, sound=sound, device=device)
+        # If device is the empty string or not specified at all, use None to send to all devices
+        # (see https://github.com/thp/urlwatch/issues/372)
+        device = self.config.get('device', None) or None
+        priority = {
+            'lowest': chump.LOWEST,
+            'low': chump.LOW,
+            'normal': chump.NORMAL,
+            'high': chump.HIGH,
+            'emergency': chump.EMERGENCY,
+        }.get(self.config.get('priority', None), chump.NORMAL)
+        msg = service.create_message(title=title, message=body, html=True, sound=sound, device=device, priority=priority)
         msg.send()
 
 
@@ -462,6 +467,9 @@ class PushbulletReport(WebServiceReporter):
     __kind__ = 'pushbullet'
 
     def web_service_get(self):
+        if Pushbullet is None:
+            raise ImportError('Python module "pushbullet" not installed')
+
         return Pushbullet(self.config['api_key'])
 
     def web_service_submit(self, service, title, body):
@@ -577,7 +585,7 @@ class TelegramReporter(TextReporter):
 
 class SlackReporter(TextReporter):
     """Custom Slack reporter"""
-    MAX_LENGTH = 4096
+    MAX_LENGTH = 40000
 
     __kind__ = 'slack'
 
@@ -614,3 +622,116 @@ class SlackReporter(TextReporter):
 
     def chunkstring(self, string, length):
         return (string[0 + i:length + i] for i in range(0, len(string), length))
+
+
+class MarkdownReporter(ReporterBase):
+    def submit(self):
+        cfg = self.report.config['report']['markdown']
+        show_details = cfg['details']
+        show_footer = cfg['footer']
+
+        if cfg['minimal']:
+            for job_state in self.report.get_filtered_job_states(self.job_states):
+                pretty_name = job_state.job.pretty_name()
+                location = job_state.job.get_location()
+                if pretty_name != location:
+                    location = '%s (%s)' % (pretty_name, location)
+                yield '* ' + ': '.join((job_state.verb.upper(), location))
+            return
+
+        summary = []
+        details = []
+        for job_state in self.report.get_filtered_job_states(self.job_states):
+            summary_part, details_part = self._format_output(job_state)
+            summary.extend(summary_part)
+            details.extend(details_part)
+
+        if summary:
+            yield from ('%d. %s' % (idx + 1, line) for idx, line in enumerate(summary))
+            yield ''
+
+        if show_details:
+            yield from details
+
+        if summary and show_footer:
+            yield from ('--- ',
+                        '%s %s, %s  ' % (urlwatch.pkgname, urlwatch.__version__, urlwatch.__copyright__),
+                        'Website: %s  ' % (urlwatch.__url__,),
+                        'watched %d URLs in %d seconds' % (len(self.job_states), self.duration.seconds))
+
+    def _format_content(self, job_state):
+        if job_state.verb == 'error':
+            return job_state.traceback.strip()
+
+        if job_state.verb == 'unchanged':
+            return job_state.old_data
+
+        if job_state.old_data in (None, job_state.new_data):
+            return None
+
+        return job_state.get_diff()
+
+    def _format_output(self, job_state):
+        summary_part = []
+        details_part = []
+
+        pretty_name = job_state.job.pretty_name()
+        location = job_state.job.get_location()
+        if pretty_name != location:
+            location = '%s (%s)' % (pretty_name, location)
+
+        pretty_summary = ': '.join((job_state.verb.upper(), pretty_name))
+        summary = ': '.join((job_state.verb.upper(), location))
+        content = self._format_content(job_state)
+
+        summary_part.append(pretty_summary)
+
+        details_part.append('### ' + summary)
+        if content is not None:
+            details_part.extend(('', '```', content, '```', ''))
+        details_part.extend(('', ''))
+
+        return summary_part, details_part
+
+
+class MatrixReporter(MarkdownReporter):
+    """Custom Matrix reporter"""
+    MAX_LENGTH = 4096
+
+    __kind__ = 'matrix'
+
+    def submit(self):
+        if matrix_client is None:
+            raise ImportError('Python module "matrix_client" not installed')
+
+        homeserver_url = self.config['homeserver']
+        access_token = self.config['access_token']
+        room_id = self.config['room_id']
+
+        body_markdown = '\n'.join(super().submit())
+
+        if not body_markdown:
+            logger.debug('Not calling Matrix API (no changes)')
+            return
+
+        if len(body_markdown) > self.MAX_LENGTH:
+            body_markdown = body_markdown[:self.MAX_LENGTH]
+
+        client_api = matrix_client.api.MatrixHttpApi(homeserver_url, access_token)
+
+        if Markdown is not None:
+            body_html = Markdown().convert(body_markdown)
+
+            client_api.send_message_event(
+                room_id,
+                "m.room.message",
+                content={
+                    "msgtype": "m.text",
+                    "format": "org.matrix.custom.html",
+                    "body": body_markdown,
+                    "formatted_body": body_html
+                }
+            )
+        else:
+            logger.debug('Not formatting as Markdown; dependency on markdown2 not met?')
+            client_api.send_message(room_id, body_markdown)
